[BASIC]
function-naming-style=snake_case
const-naming-style=snake_case
attr-naming-style=snake_case
argument-naming-style=snake_case
module-naming-style=snake_case
method-naming-style=snake_case
variable-naming-style=snake_case
class-attribute-naming-style=snake_case
inlinevar-naming-style=snake_case
class-naming-style=snake_case

include-naming-hint=yes
